;;;			     mew-mime.el
;;;
;;;		Copyright (C) 1997  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazu Yamamoto <Kazu@Mew.org>
;;; Created: Mar 23, 1997
;;; Revised: Oct 25, 1997
;;;

(defconst mew-mime-version "mew-mime.el version 0.10")

(require 'mew)

(defvar mew-eof-string nil
  "*A value inserted on the end of message buffer if non-nil.")

(defvar mew-mime-content-type-text-list
  '("Text/Plain" "Text/Html"
    "Application/Postscript"
    "Message/Rfc822")
  "Content-Type: list for text which are treated as text when encoded"
  )

(defvar mew-mime-content-type-multipart-list
  '("Multipart/Mixed" "Multipart/Alternative"
    ;;"Multipart/Digest" "Multipart/Parallel"
    )
  "Candidate of 'Content-Type: Multipart' when CT: is changed in draft buffer."
  )

(defvar mew-mime-content-type-binary-list
  '("Application/Octet-Stream"
    "Image/Jpeg" "Image/Gif" "Image/x-xwd" "Image/x-xbm" "Image/x-bmp"
    "Audio/Basic" "Video/Mpeg"
    )
  "Content-Type: list to save as binary file in mew-summary-save on Mule"
  )

(defvar mew-mime-content-type-others-list
  '("Message/Rfc822" "Message/External-body")
  )

(defvar mew-mime-content-type-list
  (append mew-mime-content-type-text-list
	  mew-mime-content-type-multipart-list
	  mew-mime-content-type-binary-list
	  mew-mime-content-type-others-list)
  "Candidate of Content-Type: when CT: is changed in draft buffer."
  )

(defvar 
 mew-mime-content-type
 (list
   (list 
    "multipart/*"
    nil
    nil
    nil
    mew-icon-multipart
    )
   (list 
    "audio/basic"
    "\\.au$"
    mew-b64
    (list mew-prog-shell (list mew-prog-shell-arg "cat - > /dev/audio") nil)
    mew-icon-audio
    )
   (list
    "image/gif"
    "\\.gif$"
    mew-b64
    (if (and mew-xemacs-p (featurep 'gif))
	'(mew-mime-image/gif () nil)
      '("xv" ("-geometry" "+0+0") t))
    mew-icon-image
    )
   (list
    "image/jpeg"
    "\\.jpe?g$"
    mew-b64
    (if (and mew-xemacs-p (featurep 'jpeg))
	'(mew-mime-image/jpeg () nil)
      '("xv" ("-geometry" "+0+0") t))
    mew-icon-image
    )
   (list
    "image/x-xwd"
    "\\.xwd$"
    mew-b64
    '("xv" ("-geometry" "+0+0") t)
    mew-icon-image
    )
   (list
    "image/x-xbm"
    "\\.xbm$"
    mew-b64
    '("xv" ("-geometry" "+0+0") t)
    mew-icon-image
    )
   (list
    "image/x-bmp"
    "\\.bmp$"
    mew-b64
    '("xv" ("-geometry" "+0+0") t)
    mew-icon-image
    )
   (list
    "image/.*"
    "^$"
    mew-b64
    '("xv" ("-geometry" "+0+0") t)
    mew-icon-image
    )
   (list
    "video/mpeg"
    "\\.mpe?g$"
    mew-b64
    '("mpeg_play" () t)
    mew-icon-video
    )
   (list
    "application/postscript"
    "\\.ps$"
    mew-qp
    '("ghostview" ("-geom" "+0+0") t)
    mew-icon-application/postscript
    )
   (list
    "application/pgp-keys"
    "\\.pka$"
    nil
    '(mew-mime-pgp-keys () nil)
    mew-icon-unknown
    )
   (list
    "application/octet-stream"
    "\\.tar\\.?g?z?$\\|\\.tgz$\\|\\.gz$\\|\\.Z$\\|\\.bin$\\|\\.pgp$"
    mew-b64
    '(mew-mime-application/octet-stream () nil)
    mew-icon-application/octet-stream
    )
   (list
    "message/external-body" 
    "\\.ext$"
    nil
    '(mew-mime-external-body () nil)
    mew-icon-message/external-body
    )
   (list
    "message/rfc822"
    "^[0-9]+$"
    nil
    '(mew-mime-message/rfc822 () nil)
    mew-icon-message/rfc822
    )
   (list
    "message/delivery-status"
    "^$"
    nil
    '(mew-mime-text/plain () nil)
    mew-icon-text
    )
   (list
    "text/richtext"
    "\\.rtf$"
    nil
    '(mew-mime-text/plain () nil)
    mew-icon-text
    )
   (list
    "text/plain"
    ".*"
    nil
    '(mew-mime-text/plain () nil)
    mew-icon-text
    )
   (list
    "text/.*"
    "^$"
    nil
    '(mew-mime-text/plain () nil)
    mew-icon-text
    )
   (list
    ".*" 
    "^$"
    nil
    '(mew-mime-application/octet-stream () nil)
    mew-icon-unknown
    )
   )
 "(content-type filename encoding (program (options ...) async) icon)"
 )

;;
(defmacro mew-file-attr (filename alist)
  (` (mew-assoc-match2 (, filename) (, alist) 1)))

(defmacro mew-type-attr (filename alist)
  (` (mew-assoc-equal (, filename) (, alist) 0)))

;; for mapcar
(defun mew-file-content (attr)
  (car attr))

(defmacro mew-file-encoding (attr)
  (` (car (cdr (cdr (, attr))))))

;;
(defmacro mew-content-attr (content alist)
  (` (mew-assoc-match2 (, content) (, alist) 0)))

(defmacro mew-content-program (attr)
  (` (car (car (cdr (cdr (cdr (, attr))))))))

(defmacro mew-content-options (attr)
  (` (car (cdr (car (cdr (cdr (cdr (, attr)))))))))


(defmacro mew-content-async (attr)
  (` (car (cdr (cdr (car (cdr (cdr (cdr (, attr))))))))))

;;
;;
;;

(defun mew-mime-start-process (program options file)
  (let ((process-connection-type mew-connection-type1) pro)
    (message "Starting %s ..." program)
    (setq pro (apply (function start-process)
		     (format "*mew %s*" program)
		     mew-buffer-tmp ;; xxx
		     program
		     (append options (list file))))
    (set-process-sentinel pro 'mew-mime-start-process-sentinel)
    (message "Sending %s ... done" program)
    (setq mew-process-file-alist (cons (cons pro file) mew-process-file-alist))
    )
  t ;; to next part
  )

(defun mew-mime-start-process-sentinel (process event)
  (let ((al (assoc process mew-process-file-alist)))
    (if (cdr al) (delete-file (cdr al)))
    (setq mew-process-file-alist (delete al mew-process-file-alist))
    ))

(defun mew-mime-call-process (program options file)
  (message "Calling %s ..." program)
  (apply (function call-process) program file nil nil options)
  (message "Calling %s ... done" program)
  t ;; to next part
  )

(defun mew-mime-image (begin end format)
  (set-buffer (mew-buffer-message))
  (let ((buffer-read-only nil))
    (message "Loading image...")
    (set-extent-end-glyph (make-extent (point-min) (point-min))
			  (make-glyph (vector 
				       format
				       :data
				       (buffer-substring 
					begin end
					(mew-current-get 'cache)))))
    (message "Loading image...done")
    ))

(defun mew-mime-image/jpeg (begin end &optional params)
  (mew-mime-image begin end 'jpeg))

(defun mew-mime-image/gif (begin end &optional params)
  (mew-mime-image begin end 'gif))

(defun mew-mime-text/plain (begin end &optional params)
  (if (> end begin)
      (save-excursion
	(set-buffer (mew-buffer-message))
	(let ((buffer-read-only nil))
	  (insert-buffer-substring (mew-current-get 'cache) begin end)
	  (if mew-eof-string
	      (progn
		(goto-char (point-max)) ;; necessary?
		(insert mew-eof-string)))
	  )
	(cond (mew-break-pages 
	       (goto-char (point-min))
	       (mew-message-narrow-to-page)
	       ))
	(set-buffer-modified-p nil) ;; xxx
	))
  )

(defun mew-mime-message/rfc822 (part)
  (let* ((hbegin (mew-syntax-get-begin part))
	 (hend   (mew-syntax-get-end   part))
	 (cache  (mew-current-get 'cache))
	 (buffer-read-only nil))
    (insert-buffer-substring cache hbegin hend)
    (if mew-decode-result (insert mew-decode-result))
    (insert "\n")
    (mew-header-arrange)
    (if (mew-syntax-singlepart-p (mew-syntax-get-part part))
	(mew-summary-display-part part nil 'non-erase)) ;; nil is single
    ;; display-part sets citation noheader
    (setq mew-message-citation 'header)
    (set-buffer-modified-p nil) ;; xxx
    ))

(defun mew-mime-application/octet-stream (begin end &optional params)
  (let ((size (- end begin))
	(buffer-read-only nil))
    (insert "######    ###   #     #    #    ######  #     #\n"
	    "#     #    #    ##    #   # #   #     #  #   #\n"
	    "#     #    #    # #   #  #   #  #     #   # #\n"
	    "######     #    #  #  # #     # ######     #\n"
	    "#     #    #    #   # # ####### #   #      #\n"
	    "#     #    #    #    ## #     # #    #     #\n"
	    "######    ###   #     # #     # #     #    #\n"
	    "\n\n"
;;	    (format "Content-Type:\t%s\n" ct)
	    (format "Size:\t\t%d bytes\n" size)
	    "\nTo save this part, type "
	    (substitute-command-keys
	     "\\<mew-summary-mode-map>\\[mew-summary-save].")
	    "\nTo display this part in Message mode, type "
	    (substitute-command-keys
	     "\\<mew-summary-mode-map>\\[mew-summary-insert].")
	    )
    ))

(defun mew-summary-insert ()
  (interactive)
  (let* ((ofld-msg (mew-current-get 'message))
	 (msg (mew-summary-message-number))
	 (part (mew-summary-part-number))
	 (buf (buffer-name)))
    (if (or msg (not part))
	(let ((mew-analysis nil))
	  (mew-summary-display))
      (unwind-protect
	  (progn
	    (mew-summary-toggle-disp-msg 'on)
	    (mew-window-configure buf 'message)
	    (set-buffer (mew-buffer-message))
	    (let* ((buffer-read-only nil)
		   (syntax (mew-cache-decode-syntax (mew-cache-hit ofld-msg)))
		   (stx (mew-syntax-get-entry-strnum syntax part))
		   (begin (mew-syntax-get-begin stx))
		   (end (mew-syntax-get-end stx)))
	      (erase-buffer)
	      (insert-buffer-substring (mew-current-get 'cache) begin end)
	      (goto-char (point-min))
	      ))
	(pop-to-buffer buf)))
    ))

(provide 'mew-mime)
